from .trainer import *
from .scheduler_callbacks import *
from .utils import *
from .package_info import *

from .trainer import __all__ as _t_all
from .scheduler_callbacks import __all__ as _s_all
from .utils import __all__ as _u_all

__all__ = _t_all + _s_all + _u_all



from torch.optim import lr_scheduler

from .trainer import Trainer, TrainerCallback
from .utils import is_divisor

__all__ = [
    'ScheduleBatchSize',
    'ScheduleLR',
    'StepLR',
]


class ScheduleBatchSize(TrainerCallback):
    def __init__(self, step: int, gamma: int = 2):
        self.step = step
        self.gamma = gamma

    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        if is_divisor(batch_num, self.step):
            trainer.batch_size *= self.gamma


class ScheduleLR(TrainerCallback):
    def __init__(self, lr_scheduler_cls, **kwargs):
        self.lr_scheduler_cls = lr_scheduler_cls
        self.kwargs = kwargs
        self.lr_scheduler = None

    def start_training(self, trainer: Trainer) -> None:
        self.lr_scheduler = self.lr_scheduler_cls(trainer.optim, **self.kwargs)
        trainer.callback_params['lrs'] = []

    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        self.lr_scheduler.step()


class StepLR(ScheduleLR):
    def __init__(self, step_size: int, gamma: float, last_epoch: int = -1, **kwargs):
        super().__init__(lr_scheduler.StepLR, step_size=step_size, gamma=gamma, last_epoch=last_epoch, **kwargs)



description = 'A minimalistic alternative to PyTorch Lightning (no batteries included)'

version_info = (0, 0, 1, 1)

__version__ = '.'.join(map(str, version_info))

__license__ = 'MIT'
__author__ = 'Vladimir Starostin'
__email__ = 'vladimir.starostin@uni-tuebingen.de'

classifiers = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
]



from typing import Tuple, Iterable, Any, Union
from collections import defaultdict

from tqdm import trange
import numpy as np

import torch
from torch.nn import Module

from .utils import is_divisor

__all__ = [
    'Trainer',
    'TrainerCallback',
    'Dataloader',
]


class Trainer(object):
    def __init__(self, model: Module, loader: 'Dataloader', lr: float, batch_size: int, **kwargs):
        self.model = model
        self.loader = loader
        self.batch_size = batch_size

        self.optim = self.configure_optimizer(lr=lr)
        self.losses = defaultdict(list)

        self.callback_params = {}
        self.lrs = []

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.init()

    def init(self):
        pass

    def train_epoch(self,
                    num_batches: int,
                    callbacks: Tuple['TrainerCallback', ...] or 'TrainerCallback' = (),
                    disable_tqdm: bool = False,
                    update_tqdm_freq: int = 10,
                    ):

        if isinstance(callbacks, TrainerCallback):
            callbacks = (callbacks, )

        callbacks = _StackedTrainerCallbacks(list(callbacks) + [self.loader])

        pbar = trange(num_batches, disable=disable_tqdm)

        callbacks.start_training(self)

        for batch_num in pbar:
            self.model.train()
            self.optim.zero_grad()
            batch_data = self.get_batch(batch_num)
            loss_dict = self.get_loss_dict(batch_data)
            loss = sum(loss_dict.values())
            loss.backward()
            self.optim.step()

            self._update_losses(loss_dict, loss)
            self.lrs.append(self.lr)

            if not disable_tqdm:
                self._update_tqdm(pbar, batch_num, update_tqdm_freq)

            break_epoch = callbacks.end_batch(self, batch_num)

            if break_epoch:
                break

        callbacks.end_training(self)

    def _update_tqdm(self, pbar, batch_num: int, update_tqdm_freq: int):
        if is_divisor(batch_num, update_tqdm_freq):
            last_loss = np.mean(self.losses['total_loss'][-10:])
            pbar.set_description(f'Loss = {last_loss:.2e}')

    def get_batch(self, batch_num: int) -> Any:
        return self.loader.get_batch(self.batch_size)

    def get_loss_dict(self, batch_data) -> dict:
        pass

    def _update_losses(self, loss_dict, loss) -> None:
        for k, v in loss_dict.items():
            self.losses[k].append(v.item())

        self.losses['total_loss'].append(loss.item())

    def configure_optimizer(self, lr: float) -> torch.optim.Optimizer:
        optim = torch.optim.AdamW(self.model.parameters(), lr)
        return optim

    @property
    def lr(self) -> float:
        return self.optim.param_groups[0]['lr']

    @lr.setter
    def lr(self, lr: float) -> None:
        self.optim.param_groups[0]['lr'] = lr


class TrainerCallback(object):
    def start_training(self, trainer: Trainer) -> None:
        pass

    def end_training(self, trainer: Trainer) -> None:
        pass

    def end_batch(self, trainer: Trainer, batch_num: int) -> Union[bool, None]:
        pass


class Dataloader(TrainerCallback):
    def get_batch(self, batch_num: int) -> Any:
        pass

    def __getitem__(self, batch_num: int) -> Any:
        return self.get_batch(batch_num)


class _StackedTrainerCallbacks(TrainerCallback):
    def __init__(self, callbacks: Iterable[TrainerCallback]):
        self.callbacks = tuple(callbacks)

    def start_training(self, trainer: Trainer) -> None:
        for c in self.callbacks:
            c.start_training(trainer)

    def end_training(self, trainer: Trainer) -> None:
        for c in self.callbacks:
            c.end_training(trainer)

    def end_batch(self, trainer: Trainer, batch_num: int) -> Union[bool, None]:
        break_epoch = False
        for c in self.callbacks:
            break_epoch += bool(c.end_batch(trainer, batch_num))
        return break_epoch



from setuptools import setup, find_packages
from pathlib import Path


def read_file(filename: str) -> str:
    with open(Path(__file__).parent / filename, mode='r', encoding='utf-8') as f:
        return f.read()


def get_package_info(package_name: str) -> dict:
    package_info_dict = {}
    with open(str(Path(__file__).parent / 'src' / package_name / 'package_info.py'), 'r') as f:
        exec(f.read(), {}, package_info_dict)

    return package_info_dict


PACKAGE_NAME = 'lighter_torch'

package_info = get_package_info(PACKAGE_NAME)

__version__ = package_info['__version__']
__license__ = package_info['__license__']
__author__ = package_info['__author__']
__email__ = package_info['__email__']

classifiers = package_info['classifiers']
description = package_info['description']

long_description = read_file('README.md')
long_description_content_type = 'text/markdown'


python_requires = '>=3.7'
install_requires = read_file('requirements.txt').splitlines()


setup(
    name=PACKAGE_NAME,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    version=__version__,
    author=__author__,
    author_email=__email__,
    license=__license__,
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    python_requires=python_requires,
    classifiers=classifiers,
    install_requires=install_requires
)
